/*
Provides functions which allow you to use the javascript type system to constuct useful assertions.
A number of Douglas Crockford's elementary javascript functions have also been reimplemented.
*/
define(['/libs/wyvern/util/errors.js',
        '/libs/wyvern/util/prototypes.js'],
    function (er) {
    var _ = {};        // Public namespace.
    var __ = {};       // Private namespace.
    /*
    An enhanced version of the typeof statement which distinguishes structurally between objects,
    arrays, and the null type.
    */
    _.typeOf = function (value) {
        var type = typeof value;
        if (type === "object") {
            // If the type has been set, use that by default.
            if (value instanceof Error) {
                type = "error";
            }
            else if (value instanceof RegExp) {
                type = "regex";
            }
            else if (value instanceof Date) {
                type = "date";
            }
            //else if (typeof value.length === "number"
            //   && value.propertyIsEnumerable("length")
            //   && typeof value.splice === "function") {
            //   type = "array";
            //}
            else if (value instanceof String) {
                type = "string";
            }
            else if (value instanceof Array) {
                type = "array";
            }
            else if (value instanceof Number) {
                type = "number"; 
            }
            else if (value instanceof Boolean) {
                type = "boolean";
            }
            else {
                type = "null";
            }
        }
        return type;
    };
    /*
    Convenience method that checks if a value is null or undefined.
    */
    _.isVoid = function (value) {
        return _.typeOf(value) in ["null", "undefined"];
    };
    /*
    Evalutes a predicate; if it is true, control is returned to the routine that called assert.
    Otherwise, an Error is thrown. If the msg is a string, an AssertionError
    */
    _.assert = function (predicate, msg, args, ctxt) {
        var criteria;
        if (_.typeOf(predicate) === "function") {
            if (_.typeOf(args) === "array" || !(_.isVoid(ctxt))) {
                // If there is a context, evaluate the predicate in that context.
                criteria = predicate.apply(
                    ctxt ? ctxt : this,
                    args
                ); 
            }
            else {
                criteria = predicate();
            }
        }
        else if (_.typeOf(predicate) === "boolean"){
            criteria = predicate; 
        }
        else  {
            throw new er.ArgumentError(
                "A boolean or a function that returns a boolean must be passed as a predicate."
            );
        }
        if (criteria) {
            return;
        }
        else if (_.typeOf(msg) === "error") {
           var err = msg;
        }
        else if (_.typeOf(msg) === "string" || _.isVoid(msg)){
            var err = new er.AssertionError(msg);
        }
        else {
            throw new er.ArgumentError(
                "The argument msg must be of type string, error, undefined, or null."
            );
        }
        throw err;
    };
    /*
    A convenience method to ensure that the type of an value is as specified else throws
    an Error. If msg is of type error, the msg is thrown. If it is a string, a TypeError is
    thrown.
    */
    _.assertTypeIs = function (value, type, name, msg) {
        _.assert(
            _.typeOf(value) in ["string", "array"],
            new er.ArgumentError("The value argument must be either a string or array.")
        );
        if (_.typeOf(type) === "string") {
            _.assert(
                _.typeOf(value) === type,
                _.typeOf(msg) === "error" ?
                    msg :
                    new er.TypeError(
                        "The argument " + name + " must be of type " + type + "."
                    )
            );
        }
        else if (_.typeOf(type) === "array") {
            _.assert(
                _.typeOf(value) in type,
                _.typeOf(msg) === "error" ?
                    msg :
                    new er.TypeError(
                        "The argument " + name + " must be one of the following types: " + type + "."
                    )
            );
        }
    };
    /*
    A convenience method to ensure that the type of an argument is as specified else throws
    an ArgumentError.
    */
    _.assertArgumentIs = function (value, type, name) {
        _.assertTypeIs(
            value, type, name,
            new er.ArgumentError(
                "The argument " + name + " must be of type: " + type + "."
            )
        );
    }
    /*
    A convenience method that checks the nullity and type of a set of arguments and throws an
    argumenent error if they do not match a specified spec.
    */
    _.assertArgumentsAre = function (args, argspec) {
        _.assertArgumentIs(args, ["object", "array"], "args");
        _.assertArgumentIs(argspec, "object", "argspec");
        var argsarr = Array.prototype.slice.call(args);
        _.assertArgumentIs(argsargs, "array", "argsarr");
        var arglen = argspec.length;
        var argkeys = Object.freeze(argspec.keys());
        var type = [];
        for (var i = 0; i < arglen; i++) {
            var varname = argkeys[i];
            var properties = argspec[varname];
            if (properties.type) {
                if (properties.required) {
                    type = type.splice(0, 2, properties.type, "null", "undefined");
                }
                else {
                    type = type.splice(0, 2, properties.type);
                }
                _.assertArgumentIs(argsarr[i], type, varname);
            }
        }
    };
    /*
    A convenience method to ensure that a value is not null or undefined. If it is, an
    AssertionError is thrown.
    */
    _.assertNotVoid = function (value, name) {
        _.assertTypeIs(value, "string", "value");
        _.assertTypeIs(name, "string", "name");
        _.assert(
            !(_.isVoid(value)),
            new er.NullValueError(
                "The value " + name + " must not be null or undefined.",
            );
        );
    };
    /*
    Checks if an object has any defined data properties (methods don't count).
    */
    _.isEmpty = function (values) {
        if (_.typeOf(values) === "object") {
            for (var i in v) {
                value = values[i];
                if (!(_.isVoid(value)) && _.typeOf(value) !== "function") {
                    return false;
                }
            }
        }
        return true;
    };
    /*
    Namespace for functions on strings.
    */
    _.string = {};
    /*
    Replaces certain characters with html entities. 
    */
    _.string.entityify = function (s) {
        // TODO:  Replace this naive implementation with a complete one.
        return str.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;").replace(/"/g, "&quot;");
    };
    /*
    TODO: document this function. 
    */
    _.string.quote = function (string) {
        // TODO: this looks broken
        var strlen = string.length; 
        var result = '"';
        var letter, i;
        for (i = 0; i < strlen; i++) {
            letter = string.charAt(i);
            if (letter >= ' ') {
                if (letter === '\\' || letter === '"') {
                    result +='\\';
                }
                result += letter;
            }
            else {
                switch (letter) {
                    case '\b': r += '\\b'; break;
                    case '\r': r += '\\r'; break;
                    case '\f': r += '\\f'; break;
                    case '\n': r += '\\n'; break;
                    case '\t': r += '\\t'; break;
                    default:
                        letter = letter.charCodeAt();
                        result += 
                            '\\u00' +
                            Math.floor(letter/16).toString(16) +
                            (letter % 16).toString(16);
                }
            }
        }
        return result + '"';
    };
    /*
    A simple string interpolator which replaces substring like "{key}" with the value associated with
    `key` in the dictionary that is passed in.
    */
    _.string.supplant = function (string) {
        return string.replace(
            /{([^{}]*)}/g,
            function (a, b) {
                var rep = string[b];
                return _.typeOf(rep) in ["string", "number"] ? rep : a;
            }
        );
    };
    /*
    Removes whitespace at the beginning and end of a string.
    */
    _.string.trim = function () {
        return s.replace(/^\s+|\s+$/g, "");
    };
    /*
    TODO: create convenience methods for setting properties.
    */
    /*
    Namespace for introspection functions on properties.
    */
    _.property = {};
    /*
    Checks if a property is enumerable.
    */
    _.property.isEnumerable = function (object, property) {
        var descriptor = Object.getOwnPropertyDescriptor(object, property);
        return descriptor ? descriptor["enumerable"] : null; 
    };
    /*
    Checks if a property is configurable.
    */
    _.property.isConfigurable = function (object, property) {
        var descriptor = Object.getOwnPropertyDescriptor(object, property);
        return descriptor ? descriptor["configurable"] : null;
    };
    /*
    Checks if a property is writable.
    */
    _.property.isWritable = function (object, property) {
        var descriptor = Object.getOwnPropertyDescriptor(object, property);
        return descriptor ? descriptor["writable"] : null;
    };
    /*
    Checks if a property has a getter function.
    */
    _.property.hasGetter = function (object, property) {
        var descriptor = Object.getOwnPropertyDescriptor(object, property);
        return descriptor ? descriptor["get"] : null;
    };
    /*
    Checks if a property has a setter function.
    */
    _.property.hasSetter =  function (object, property) {
        var descriptor = Object.getOwnPropertyDescriptor(object, property);
        return descriptor ? descriptor["set"] : null;
    };
    return _;
});
