/*
Allows you to setup a game and wraps common functionality in utility methods which
set up WebGl, create scenes, create and position cameras, & add lighting. Also
provides methods for populating the scene with entities associated with a model,
events and data.
*/
define(['/libs/wyvern/deps/lodash.js',
        '/libs/wyvern/util/introspect.js',
        '/libs/wyvern/util/logger.js'],
    function (us, is, lo) {
    var _ = {};        // Public namespace.
    var __ = {};       // Private namespace.
    /*
    Serves as a registry of messages to functions. When a message is published, if
    any functions are registered for that message, those functions are invoked.
    */
    _.Pub = function () {
        this.msgs = {};
    };
    _.Pub.fn = _.Pub.prototype;
    /*
    Adds a subscriber function to a message type.
    */
    _.Pub.fn.sub = function (msg, sub, fn)  {
        var subs = this.msgs[msg];
        if (subs && this.findSub(subs, sub) != -1) {
            return;
        }
        else {
            this.msgs[msg] = [{
                sub: sub,
                fn: fn
            }];
        }
    };
    /*
    Removes a subscriber function from a message type.
    */
    _.Pub.fn.unsub = function (msg, sub, fn) {
        if (sub) {
            var subs = this.msgs[msg];
            var i = this.findSub(subs, sub, fn);
            if (subs && i != -1) {
                this.msgs[msg].splice(i, 1);
            }
        }
        else {
            delete this.msgs[msg];
        }
    };
    /*
    Publishes a message, calling all functions that subscribe to that message with the
    arguments specified by the parameters passed after `msg`.
    */
    // TODO: this is a stupid way to write pub-sub unless all subscribers take a context
    // dictionary as their only argument.
    _.Pub.fn.pub = function (msg) {
        var subs = this.msgs[msg];
        if (subs) {
            for (var i = 0; i < subs.length; i++) {
                var args = [];
                for (var j = 1; j < arguments.length; j++) {
                    args.push(arguments[j]);
                }
                subs[i].fn.apply(subs[i].sub, args);
            }
        }
    };
    /*
    Returns the index of a subscriber in a list of subscribers.
    */
    _.Pub.fn.findSub = function (subs, sub) {
        for (var i = 0; i < subs.length; i++) {
            if (subs[i] == sub) {
                return i;
            }
        }
        return -1;
    };
    /*
    Returns a new Game object which behaves as a publisher and declares all of its fields.
    */
    _.Game = function () {
        _.Pub.call(this);
        this.renderer = null;
        this.scene = null;
        this.camera = null;
        this.entities = [];
        _.Browser.fixStyles();
    };
    _.Game.fn = _.Game.prototype =  new _.Pub();
    /*
    Initializes the game.
    */
    _.Game.fn.init = function (param) {
        var param = param || {};
        this.container = param.container;
        this.canvas = param.canvas;
        console.log("Game.fn.init -> param", param);
        console.log("Game.fn.init -> container", this.container);
        console.log("Game.fn.init -> canvas", this.canvas);
        
        // Create the Three.js renderer, add it to the div
        var renderer = new THREE.WebGLRenderer({antialias: true, canvas: this.canvas});
        // Sets the canvas to the rederer's domElement
        this.canvas = renderer.domElement;
        renderer.setSize(this.container.offsetWidth, this.container.offsetHeight);
        console.log("Game.fn.init -> renderer", renderer);
        console.log("Game.fn.init -> renderer.domElement", renderer.domElement);
        this.container.appendChild(this.canvas);

        // Put in a camera at a good default location
        var camera = new THREE.PerspectiveCamera(45, this.container.offsetWidth / this.container.offsetHeight, 1, 10000);
        camera.position.set(0, 0, 3.3333);
        
        // Create a new Three.js scene
        var scene = new THREE.Scene();
        scene.add(new THREE.AmbientLight(0x505050));
        scene.data = this;        
        scene.add(camera);

        // Create a root object to contain all other scene objects
        var root = new THREE.Object3D();
        scene.add(root);

        // Create a projector to handle picking
        var projector = new THREE.Projector();
        
        // Save a few things
        this.renderer = renderer;
        this.scene = scene;
        this.camera = camera;
        this.projector = projector;
        this.root = root;

        // Bind event handlers
        this.mouseHandlers();
        this.keyboardHandlers();
        this.domHandlers();
    };
    /*
    Animation loop which keeps rendering frames of the game and updating the state of
    the entities of which its comprised.
    */
    _.Game.fn.run = function () {
        this.update();
        this.renderer.render(this.scene, this.camera);
        requestAnimationFrame(function() {
            this.run();
        });
    };
    /*
    Renders all game entities.
    */
    _.Game.fn.draw = function () {
        for (var i = 0; i < this.entities.length; i++) {
            this.entities[i].draw();
        }
    };
    /*
    Adds an entity to the game and its model to the scene.
    */
    _.Game.fn.addEntity = function (en) {
        this.entities.push(en);
        console.log("Game.fn.addEntity -> root: ", this.root);
        console.log("Game.fn.addEntity -> en: ", en);
        if (en.obj3D) {
            this.root.add(en.obj3D);
        }
    };
    /*
    Removes an entity from the game and its model from the scene.
    */
    _.Game.fn.delEntity = function (en) {
        var i = this.entities.indexOf(en);
        if (i != -1) {
            this.entities.splice(i, 1);
            if (en.obj3D) {
                this.root.remove(en.obj3D);
            }
        }
    };
    /*
    Handler for mouse events.
    */
    _.Game.fn.mouseHandlers = function () {
        var that = this;
        var canvas = that.canvas;
        console.log("Game.fn.mouseHandlers -> that", that);
        console.log("Game.fn.mouseHandlers -> canvas", that.canvas);
        canvas.addEventListener('mousemove',
            function (e) {
                that.mouse.move(that, e);
            },
        false);
        canvas.addEventListener('mousedown',
            function (e) {
                that.mouse.click(that, e);
            },
        false);
        canvas.addEventListener('mouseup',
            function (e) {
                that.mouse.unclick(that, e);
            },
        false);
        $(canvas).mousewheel(
            function (e) {
                that.mouse.scroll(that, e);
            }
        );
        that.mouse.over = null;
        that.mouse.clicked = null;
    };
    /*
    Stores mapping of common keys to character codes.
    */
    _.KEYCODES = Object.freeze({
        LEFT: 37,
        UP: 38,
        RIGHT: 39,
        DOWN: 40
    });
    /*
    Handler for keyboard events.
    */
    _.Game.fn.keyboardHandlers = function () {
        var that = this;
        var canvas = that.canvas;
        console.log("Game.fn.keyboardHandlers -> canvas", canvas);
        canvas.addEventListener('keydown', that.keyboard.down, false);
        canvas.addEventListener('keyup', that.keyboard.up, false);
        canvas.addEventListener('keypress', that.keyboard.press, false);
        // Focus via tab cycling
        canvas.setAttribute('tabindex', 1);
        canvas.style.outline = 'none';
    };
    /*
    Handler for dom events. 
    */
    _.Game.fn.domHandlers = function () {
        var that = this;
        window.addEventListener("resize",
            function (e) {
                that.window.resize(that, e);
            },
        false);
    };
    /*
    Stores entities that are part of the game.
    */
    _.Entities = {};
    /*
    TODO: document this function.
    */
    _.Entities.fromIntersected = function (object, point, normal) {
        if (object.data) {
            return {
                object: object.data,
                point: point,
                normal: normal
            };
        }
        else if (object.parent) {
            return _.Entities.fromIntersected(object.parent, point, normal);
        }
        else {
            return {
                object: null,
                point: null,
                normal: null
            };
        }
    };
    /*
    TODO: document this function.
    */
    _.Entities.fromMouse = function (game, pagex, pagey) {
        //Translate page coords to element coords
        console.log("Entities.fromMouse -> game", game);
        console.log("Entities.fromMouse -> game.canvas", game.canvas);
        var offset = $(game.canvas).offset();
        console.log("Entities.fromMouse -> offset", offset);
        var eltx = pagex - offset.left;
        var elty = pagey - offset.top;

        //Translate client coords into viewport x,y
        var vpx = (eltx / game.container.offsetWidth) * 2 - 1;
        var vpy = (elty / game.container.offsetHeight) * 2 + 1;

        var vector = new THREE.Vector3(vpx, vpy, 0.5);
        game.projector.unprojectVector(vector, game.camera);
        // var ray = new THREE.Ray(game.camera.position, vector.sub(game.camera.position).normalize());
        var raycaster = new THREE.Raycaster(game.camera.position, vector.sub(game.camera.position).normalize());
        console.log("Entities.fromMouse -> game.camera.position", game.camera.position);

        //TODO: recursive ray casting is set to true. Check if this is necessary.
        var intersects = raycaster.intersectObject(game.scene, true);

        if (intersects.length > 0) {
            var i = 0;
            console.log("Entities.fromMouse -> intersects", intersects);
            console.log("Entities.fromMouse -> intersects.length", intersects.length);
            findVisibleEntity:
            while (true) {
                console.log("Entities.fromMouse:findVisibleEntity -> i", i);
                if (intersects[i].object.visible){
                    break;
                }
                i++;
            }
            var intersected = intersects[i];
            var mat = new THREE.Matrix4().getInverse(intersected.object.matrixWorld);
            return _.Entities.fromIntersected(intersected.object, intersected.point, intersected.face.normal);
        }
        else {
            return {
                object: null,
                point: null,
                normal: null
            };
        }
    };
    // Route mouse events
    // TODO: Make sure this eventually implements the chain-of-responsibility pattern.
    // TODO: Check that handlers execute with the right context.
    _.Game.fn.mouse = {};
    /*
    TODO: document this function.
    */
    _.Game.fn.mouse.move = function (game, e) {
        e.preventDefault();
        if (this.clicked && this.clicked.handlers && this.clicked.handlers.mouse && this.clicked.handlers.mouse.move) {
            var hitpoint = null;
            var hitnormal = null;
            console.log("Game.mouse.move -> game.canvas", game.canvas);
            var intersected = _.Entities.fromMouse(game, e.pageX, e.pageY);
            if (intersected.entity == this.clicked) {
                hitpoint = intersected.point;
                hitnormal = intersected.normal;
            }
            this.clicked.handlers.mouse.move(e.pageX, e.pageY, hitpoint, hitnormal);
        }
        else {
            var handled = false;
            var lastOver = this.over;
            console.log("Game.mouse.move -> game.canvas", game.canvas);
            var intersected = _.Entities.fromMouse(game, e.pageX, e.pageY);
            this.over = intersected.entity;
            if (this.over != lastOver) {
                if (lastOver)  {
                    this.container.style.cursor = "auto";
                    if (lastOver.handlers.mouse.out) {
                        lastOver.handlers.mouse.out(e.pageX, e.pageY);
                    }
                }
                if (this.over) {
                    if (this.over.overCursor) {
                        game.container.style.cursor = this.over.overCursor;
                    }
                    if (this.over.handlers && this.over.handlers.mouse && this.over.handlers.mouse.over) {
                        this.over.handlers.mouse.over(e.pageX, e.pageY);
                    }
                }
                handled = true;
            }
        }
        if (!handled && this.handlers && this.handlers.mouse && this.handlers.mouse.move) {
            this.handlers.mouse.move(e.pageX, e.pageY);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.mouse.click = function (game, e) {
        e.preventDefault();
        var handled = false;
        console.log("Game.mouse.click -> this.canvas", game.canvas);
        var intersected = _.Entities.fromMouse(game, e.pageX, e.pageY);
        if (intersected.entity) {
            if (intersected.entity.handlers.mouse.click) {
                intersected.entity.handlers.mouse.click(e.pageX, e.pageY, intersected.point, intersected.normal);
                this.clicked = intersected.entity;
                handled = true;
            }
        }
        if (!handled && this.handlers && this.handlers.mouse && this.handers.mouse.click) {
            this.handlers.mouse.click(e.pageX, e.pageY);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.mouse.unclick = function (game, e) {
        e.preventDefault();
        var handled = false;
        var intersected = _.Entities.fromMouse(game, e.pageX, e.pageY);
        if (intersected.entity) {
            if (intersected.entity.handlers.mouse.unclick) {
                intersected.entity.handlers.mouse.unclick(e.pageX, e.pageY, intersected.point, intersected.normal);
                handled = true;
            }
        }
        if (!handled && this.handlers && this.handlers.mouse && this.handlers.mouse.unclick) {
            this.handlers.mouse.unlick(e.pageX, e.pageY);
        }
        this.clicked = null;
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.mouse.scroll = function (game, e, d) {
        e.preventDefault();
        if (this.handlers && this.handlers.mouse && this.handlers.mouse.scroll) {
            this.handlers.mouse.scroll(delta);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.keyboard = {};
    /*
    TODO: document this function.
    */
    _.Game.fn.keyboard.down = function (e) {
        e.preventDefault();
        if (this.handlers && this.handlers.mouse && this.handlers.keyboard.down) {
            this.handlers.keyboard.down(e.keyCode, e.charCode);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.keyboard.up = function (e) {
        e.preventDefault();
        if (this.handlers && this.handlers.mouse && this.handlers.keyboard.up) {
            this.handler.keyboard.up(e.keyCode, e.charCode);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.keyboard.press = function (e) {
        e.preventDefault();
        if (this.handlers && this.handlers.mouse && this.handlers.keyboard.press) {
            this.handlers.keyboard.press(e.keyCode, e.charCode);
        }
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.window = {};
    /*
    TODO: document this function.
    */
    _.Game.fn.window.resize = function (game, e) {
        game.renderer.setSize(game.container.offsetWidth, game.container.offsetHeight);
        game.camera.aspect = game.container.offsetWidth / game.container.offsetHeight;
        game.camera.updateProjectionMatrix;
    };
    /*
    TODO: document this function.
    */
    _.Game.fn.focus = function () {
        console.log("Game.fn.focus -> this.canvas", this.canvas);
        if (this.canvas) {
            this.canvas.focus();
        }
    };

    /*
    Represents an entity within the game.
    */
    _.Entity = function () {
        _.Pub.call(this);
        this.obj3D = null;
        this.children = [];
    };
    _.Entity.fn = _.Entity.prototype = new _.Pub;
    _.Entity.fn.init = function () {};
    _.Entity.fn.update = function () {
        this.updateChildren();
    };
    _.Entity.fn.setPosition = function (x, y, z) {
        if (this.object3D) {
            this.obj3D.position.set(x, y, z);
        }
    };
    _.Entity.fn.setScale = function (x, y, z) {
        if (this.obj3D) {
            this.obj3D.scale.set(x, y, z);
        }
    };
    _.Entity.fn.update = function () {
        for (var i = 0; i < this.children.length; i++) {
            this.children[i].update();
        }
    };
    _.Entity.fn.setObj3D = function (obj3D) {
        obj3D.data = this;
        this.obj3D = obj3D;
    };
    _.Entity.fn.addChild = function (child) {
        this.children.push(child);
        if (child.obj3d) {
            this.obj3D.add(child.obj3D);
        }
    };
    _.Entity.fn.removeChild = function (chid) {
        var childIdx = this.children.indexof(child) ;
        if (childIdx != -1) {
            this.children.splice(childIdx, 1);
            if (child.obj3D) {
                this.obj3D.remove(child.obj3D);
            }
        }
    };
    _.Entity.fn.getScene = function () {
        var obj;
        if (this.obj3D) {
            obj = this.obj3D;
            while (obj.parent) {
                obj = obj.parent;
            }
        }
        return obj;
    };
    _.Entity.fn.getGame = function () {
        var scene = this.getScene();
        return scene ? scene.data : null;
    };
    _.Browser = {};
    // TODO: generalize this function to handle any set of game stylesheets
    _.Browser.fixStyles = function () {
        StyleFix.link($("link#gameStyle"));
    };
    return _;
});
